#ifndef __KLNK_PROTOCOL_STRUCTS_H__
#define __KLNK_PROTOCOL_STRUCTS_H__

#include <stdint.h>
#include <stddef.h>
#include <memory>
#include "klink.h"
#include "msgpack.h"
#include <string>
#include <vector>
#include <map>

//Forward declare.
struct msgpack_object;
struct msgpack_packer;

namespace kilolink {

//--------------------------------------------------------
// MessagePack variant
//--------------------------------------------------------

class message_variant_t {
friend class packet_t;
public:
	enum type_t {
		T_NIL                  = 0x00,
		T_BOOLEAN              = 0x01,
		T_POSITIVE_INTEGER     = 0x02,
		T_NEGATIVE_INTEGER     = 0x03,
		T_FLOAT32              = 0x0a,
		T_FLOAT64              = 0x04,
		T_FLOAT                = 0x04,
		T_STR                  = 0x05,
		T_ARRAY                = 0x06,
		T_MAP                  = 0x07,
		T_BIN                  = 0x08,
		T_EXT                  = 0x09
	};
	
	typedef std::vector<message_variant_t> array_t;
	typedef std::multimap<message_variant_t, message_variant_t> map_t;

public:
	message_variant_t();
	message_variant_t(bool x);
	message_variant_t(uint64_t x);
	message_variant_t(int64_t x);
	message_variant_t(float x);
	message_variant_t(double x);
	message_variant_t(const std::string& x);
	message_variant_t(const char* x, int len = -1);
	message_variant_t(const uint8_t* x, size_t len);
	message_variant_t(const message_variant_t& from);
	message_variant_t(const msgpack_object& from);
	~message_variant_t();

public:
	//TODO: 暂只能考虑简单类型的直接赋值
	void assign_nil();
	void assign(bool x);
	void assign(uint64_t x);
	void assign(int64_t x);
	void assign(float x);
	void assign(double x);
	void assign(const std::string& x);
	void assign(const char* x, int len = -1);
	void assign(const uint8_t* x, size_t len);
	void assign(const message_variant_t& x);
	void assign(const msgpack_object& from);

	inline message_variant_t& operator=(bool x) {
		assign(x); return *this;
	}

	inline message_variant_t& operator=(int64_t x) {
		assign(x); return *this;
	}

	inline message_variant_t& operator=(uint64_t x) {
		assign(x); return *this;
	}

	inline message_variant_t& operator=(float x) {
		assign(x); return *this;
	}

	inline message_variant_t& operator=(double x) {
		assign(x); return *this;
	}

	inline message_variant_t& operator=(const std::string& x) {
		assign(x); return *this;
	}

	inline message_variant_t& operator=(const char* x) {
		assign(x, -1); return *this;
	}

	inline message_variant_t& operator=(const message_variant_t& x) {
		assign(x); return *this;
	}

	inline message_variant_t& operator=(const msgpack_object& x) {
		assign(x); return *this;
	}

	message_variant_t::type_t type();

	inline bool is_nil() { return type() == T_NIL; }
	inline bool is_boolean() { return type() == T_BOOLEAN; }
	inline bool is_integer() { return type() == T_POSITIVE_INTEGER || type() == T_NEGATIVE_INTEGER; }
	inline bool is_positive_integer() { return type() == T_POSITIVE_INTEGER; }
	inline bool is_negative_integer() { return type() == T_NEGATIVE_INTEGER; }
	inline bool is_number() { return type() == T_FLOAT32 || type() == T_FLOAT64; }
	inline bool is_float32() { return type() == T_FLOAT32; }
	inline bool is_float64() { return type() == T_FLOAT64; }
	inline bool is_double() { return type() == T_FLOAT64; }
	inline bool is_string() { return type() == T_STR; }
	inline bool is_array() { return type() == T_ARRAY; }
	inline bool is_map() { return type() == T_MAP; }
	inline bool is_binary() { return type() == T_BIN; }
	inline bool is_ext() { return type() == T_EXT; }

	bool operator== (const message_variant_t& other) const;
	bool operator== (bool x) const;
	bool operator== (int64_t x) const;
	bool operator== (uint64_t x) const;
	bool operator== (float x) const;
	bool operator== (double x) const;
	bool operator== (const char* x) const;
	bool operator== (const std::string& x) const;
	bool operator== (const msgpack_object& x) const;
	
	bool operator< (const message_variant_t& other) const;

	template<class T>
	inline T cast(void) { return (T)0; }

	bool get_boolean(void);
	int64_t get_int64(void);
	uint64_t get_uint64(void);
	float get_float(void);
	double get_double(void);
	const char* get_cstr(size_t& size);
	std::string get_string(void);

	const uint8_t* get_binary(size_t& size);
	
	int get_array(array_t& ret_arr);

	//注意：这里的map不是真的map，只是由vector所管理的一系列pair
	int get_map(map_t& ret_map);

protected:
	uint8_t*	m_temp_buf = nullptr;
	size_t		m_buf_size = 0;
	uint64_t    m_object_stub[4]; //4*8 = 32个字节足以存储 msgpack_object 对象。
};

template<> inline bool message_variant_t::cast<bool>(void) {
	return get_boolean();
}

template<> inline int64_t message_variant_t::cast<int64_t>(void) {
	return get_int64();
}

template<> inline int32_t message_variant_t::cast<int32_t>(void) {
	return (int32_t)get_int64();
}

template<> inline int16_t message_variant_t::cast<int16_t>(void) {
	return (int16_t)get_int64();
}

template<> inline int8_t message_variant_t::cast<int8_t>(void) {
	return (int8_t)get_int64();
}

template<> inline uint64_t message_variant_t::cast<uint64_t>(void) {
	return get_uint64();
}

template<> inline uint32_t message_variant_t::cast<uint32_t>(void) {
	return (uint32_t)get_uint64();
}

template<> inline uint16_t message_variant_t::cast<uint16_t>(void) {
	return (uint16_t)get_uint64();
}

template<> inline uint8_t message_variant_t::cast<uint8_t>(void) {
	return (uint8_t)get_uint64();
}

template<> inline float message_variant_t::cast<float>(void) {
	return get_float();
}

template<> inline double message_variant_t::cast<double>(void) {
	return get_double();
}

//注意：这是危险操作，不建议使用！
template<> inline const char* message_variant_t::cast<const char*>(void) {
	size_t sz;
	return get_cstr(sz);
}

template<> inline std::string message_variant_t::cast<std::string>(void) {
	return get_string();
}

//--------------------------------------------------------
//KILOLINK Packet
//--------------------------------------------------------
class packet_t {
public:
	//如果构造时不指定size, 则一开始不会分配内存
	packet_t(size_t size = 0); 

	//将一个已存在的buffer(from和size)包装成packet。
	//reuse表示是否直接利用from（缺省不是）。这其中的意思是，
	//默认不reuse，则构造时会复制from中的内容到packet中。而
	//reuse则直接使用from的内存。请注意：
	//  1. reuse的内存必须是使用 malloc/realloc 分配的
	//  2. reuse后，packet将会接管你的内存，所以你不可以自行
	//     释放内存。
	packet_t(void* from, size_t size, bool reuse = false);

	~packet_t();

	//alloc分配内存。如果已经分配，再次指定size时：
	//将使用realloc(...)重新分配
	//但是注意：一旦调用add_to_klnk而且成功过，则不再允许alloc
	inline uint8_t* alloc(size_t size) {
		if(size <= 0) return m_buffer;
		if(m_added_to_klnk) {
			//已经added to klnk，不再分配
			return m_buffer;
		}
		else if(!m_buffer) {
			m_buffer = (uint8_t*)malloc(size);
			if(m_buffer)
				m_max_size = size;
			return m_buffer;
		}
		else {
			uint8_t* new_ptr = (uint8_t*)realloc(m_buffer, size);
			if(new_ptr) {
				m_buffer = new_ptr;
				m_max_size = size;
			}
			return m_buffer;
		}
	}

	//释放alloc的内存。但是如果add_to_klnk成功，则内存不会自
	//动释放，而是交由KLNK底层去处理。
	inline bool free(void) {
		if(m_added_to_klnk) return true;
		else if(m_buffer) {
			::free(m_buffer);
			m_buffer = nullptr;
			m_max_size = 0;
			m_size = 0;
			m_size_size = 0;
			m_size_cur_bytes = 0;
			return true;
		}
		return false;
	}

	//返回分配的最大内存大小
	inline size_t   get_max_size(void) const {return m_max_size;}

	//获取packet的buffer地址。如果未分配，返回nullptr;
	//注意：如果 has_added_to_klnk() == true，则*不建议*直接
	//访问该内存，因为这很有可能是危险的非法内存访问。
	inline uint8_t* get(void) const { return m_buffer; }
	inline uint8_t* get_buffer(void) const { return m_buffer; } //二者相同，但为了防止与智能指针的get引起混乱，增加一个get_buffer

	//get_size/set_size 用于获取/设定packet的实际数据大小。注意
	//alloc()是分配的最大内存空间，实际的数据大小必须通过set_size指定
	inline size_t   get_size(void) const {return m_size;}
	inline void     set_size(size_t size, bool re_alloc = false) 
	{
		if(size <= m_max_size) {
			m_size = size;
			if(re_alloc && size > 0 && ((size+7)&0xfffffff8) < m_max_size) 
				alloc((size+7)&0xfffffff8);
		}
	}

	//set_fin/get_fin 用于设定发送这个packet时，是否设定FIN标志。
	//这与add_to_klnk时 'set_fin' 标志的关系是：
	//如果set_fin(1)，则无论add_to_klnk的set_fin标志是什么，都会发送FIN标志；
	//如果set_fin(0)，则由add_to_klnk的set_fin标志决定发送时是否FIN。
	inline void     set_fin(int fin) { m_fin = !!fin; }
	inline int      get_fin(void) const { return m_fin; }

	inline bool		is_valid(void) const { return (m_size > 0 || m_fin != 0); }
	inline void     clear(void) {m_size = 0; m_size_cur_bytes = 0; m_size_size = 0; m_fin=0; }

	//packet的检查/追加模式。
	//packet在stream上的发送，遵循如下格式：
	// [VAR-LEN][....packet数据...]
	//VAR-LEN是可变长度的整数格式(见k_utils.h)，表示当前这个包的
	//字节数。在检查/追加模式下，会首先检查VAR-LEN，然后持续地
	//接收数据直到满足VAR-LEN为止。
	//
	//used_size返回从from_buf中利用了多个字节的数据。
	//返回值 true 表示packet已经完整收取；false表示还没有完全收取。
	bool   check_and_push(const uint8_t* from_buf, size_t buf_size, size_t& used_size);

public:
	ssize_t  add_to_klnk(klnk_connection_t conn, klnk_stream_id_t stream_id, int set_fin = 0);
	bool     has_added_to_klnk(void) {	return m_added_to_klnk;	}

public:
	static int msgpack_write(void* data, const char* buf, size_t len) {
		packet_t* self = (packet_t*)data;
		if(self->append_buf((uint8_t*)buf, len) >= 0) return 0;
		else return -1;
	}

// protected:
	int append_buf(uint8_t* buf, size_t len);
	//inline struct msgpack_packer* get_packer();

public: //pack部分
	ssize_t pack_int8(int8_t x);
	//overload pack()
	inline ssize_t pack(int8_t x) { return pack_int8(x); }

	ssize_t pack_int16(int16_t x);
	//overload pack()
	inline ssize_t pack(int16_t x) { return pack_int16(x); }

	ssize_t pack_int32(int32_t x);
	//overload pack()
	inline ssize_t pack(int32_t x) { return pack_int32(x); }

	ssize_t pack_int64(int64_t x);
	//overload pack()
	inline ssize_t pack(int64_t x) { return pack_int64(x); }

	ssize_t pack_uint8(uint8_t x);
	//overload pack()
	inline ssize_t pack(uint8_t x) { return pack_uint8(x); }

	ssize_t pack_uint16(uint16_t x);
	//overload pack()
	inline ssize_t pack(uint16_t x) { return pack_uint16(x); }

	ssize_t pack_uint32(uint32_t x);
	//overload pack()
	inline ssize_t pack(uint32_t x) { return pack_uint32(x); }

	ssize_t pack_uint64(uint64_t x);
	//overload pack()
	inline ssize_t pack(uint64_t x) { return pack_uint64(x); }

	ssize_t pack_nil(void);
	//overload pack()
	inline ssize_t pack(void) { return pack_nil(); }

	ssize_t pack_bool(bool x);
	//overload pack()
	inline ssize_t pack(bool x) { return pack_bool(x); }

	ssize_t pack_float(float x);
	//overload pack()
	inline ssize_t pack(float x) { return pack_float(x); }

	ssize_t pack_double(double x);
	//overload pack()
	inline ssize_t pack(double x) { return pack_double(x); }

	ssize_t pack_string(const std::string& x);
	//overload pack()
	inline ssize_t pack(const std::string& x) { return pack_string(x); }

	ssize_t pack_string(const char* x, int len = -1); //如果len不指定，则自动取strlen(x)
	//overload pack()
	inline ssize_t pack(const char* x, int len = -1) { return pack_string(x, len); }

	ssize_t pack_bin(const uint8_t* buf, size_t len);
	//overload pack()
	inline ssize_t pack(const uint8_t* buf, size_t len) { return pack_bin(buf, len); }

	ssize_t pack_variant(const message_variant_t& v);
	//overload pack()
	inline ssize_t pack(const message_variant_t& v) { return pack_variant(v); }

	ssize_t pack_object(const msgpack_object& o);

	//(msgpack)原始方式pack array: 随后pack指定size个数的内容
	ssize_t pack_array_raw(int array_size);

	//(msgpack)原始方式pack map: 随后pack指定size*2个数的内容 <key,val>
	ssize_t pack_map_raw(int map_size);

	//我们自定义方式的 pack 数组，当然不是万能的 -
	//由于 message_variant_t 构造嵌套的复杂结构很麻烦（没有实现），所以variant只能是简单类型和字符串、二进制
	//所以如果是自行构建的array，里面的元素只会有简单类型和字符串、二进制。
	//当然，如果你的message_variant_t来自于msgpack unpack后的对象，那就不受限制。
	ssize_t pack_array( const message_variant_t::array_t& arr );

	//我们自定义方式的 pack map，当然不是万能的 -
	//由于 message_variant_t 构造嵌套的复杂结构很麻烦（没有实现），所以variant只能是简单类型和字符串、二进制
	//所以如果是自行构建的map，里面每一个pair 的<key(first),value(second)> 都
	//会是简单类型和字符串、二进制。
	//当然，如果message_variant_t来自于msgpack unpack后的对象，那就不受限制。
	ssize_t pack_map( const message_variant_t::map_t& map );

public:
	//使用智能指针以免时时惦记释放问题
	class base_unpacker {
	public:
		base_unpacker() {}
		virtual ~base_unpacker() {}
	};

	typedef std::shared_ptr<base_unpacker> ptr_unpacker;
public:
	//新建一个unpacker
	ptr_unpacker new_unpacker();
	//重置 unpacker - 使其可以从头重新解析
	void reset_unpacker(ptr_unpacker upk);

	//解析(unpack)下一个对象。如果成功，返回>0；失败则返回 <=0
	//返回对象存储在ret中。
	int unpack_next(ptr_unpacker upk, message_variant_t& ret);

	//简易版unpack: 取下一个对象。如果失败，返回variant对象.is_nil() == true
	message_variant_t unpack_next(ptr_unpacker upk);

protected:
	size_t	m_max_size = 0;
	size_t	m_size = 0;
	uint8_t m_size_hdr[8];
	size_t	m_size_size = 0;
	size_t  m_size_cur_bytes = 0;
	uint8_t	*m_buffer = nullptr;
	bool	m_added_to_klnk = false;
	int		m_fin = 0;
	struct msgpack_packer* m_packer = nullptr;
};

typedef std::shared_ptr<packet_t>	ptr_packet_t;

}

#endif
