#ifndef __KLNK_CPP_H__
#define __KLNK_CPP_H__

#include <sys/types.h>
#include <sys/stat.h>
#include <string>
#include <thread>
#include <memory>
#include <functional>
#include <mutex>
#include <atomic>
#include "klink.h"
#include "kloop.h"
#include "k_proto.hpp"
#include <map>
#include <queue>
#include <list>

namespace kilolink {

//--------------------------------------------------------
//Link mode
//--------------------------------------------------------
enum link_mode_t
{
	mode_client = 0,
	mode_server
};

#define INVALID_STREAM_ID		(0xffffffffffffffffULL)
//--------------------------------------------------------
//创建KILOLINK的选项
//--------------------------------------------------------
struct options_t
{
	//全局参数
	int max_connections = 1024;		   //最大允许的连接数
	int max_cache = 1024*1024;         //最大rx/tx cache大小??
	std::string cc_alg = "bbr";		   //拥塞控制算法, 如果为NULL, 默认选择"bbr" - "reno", "cubic", "dcubic", "fast", "bbr"
	bool disable_mp = false;		   //true=禁止multi-path（在某些特定情况下）; false=缺省默认允许multi-path.
	int mtu = 1400;					   //指定lower-layer的UDP mtu大小
	int klnk_mode = (int)mode_client;  //指定模式，1服务端模式，0客户端模式 (见link_mode_t)
	std::string default_alpn = "klnk"; //客户端的alpn需和服务端一致才能连接成功
	unsigned int state_interval = 500; //链路状态查询间隔，单位毫秒，0表示不查询；客户端的话，会每次把查询到的链路状态发送到服务器

	//以下为服务端模式参数
	std::string cert_file = ""; //证书文件
	std::string key_file = "";	//Key文件
	std::string root_crt = "";	//根证书文件位置（注意，如果root_crt为NULL，实现的时候要调用picoquic_set_null_verifier）
	int listen_port = 14002;	//监听端口

	//以下为客户端模式参数
	std::string type = "";				//客户端的设备类型 (不能超过12字节)
	std::string serial = "";			//客户端的序列号 (不能超过20字节)
	std::string default_path = "";		//默认链路，建议使用网卡的名称，如eth0 / wwan0
	std::string ticket_store_file = ""; //
	std::string token_store_file = "";	//
	int cipher = 10;					//加密算法，10 不加密 | TODO: 其它定义
	std::string server_addr = "";		//服务器地址
	int server_port = 14002;			//服务器端口
	unsigned int transport_calculate_interval = 0; //传输通道统计时间，单位微秒
};

//--------------------------------------------------------
//路径状态信息
//--------------------------------------------------------
struct path_state_t
{
	//网卡名称，client_path_add 时传入的参数
	std::string	ifname = "";
	
	//链路状态 IDLE/CONNECTING/CONNECTED/BROKEN/DEAD/UNKNOWN
	std::string	state = "";
	
	uint64_t rtt = 0;						//链路的rtt时间，单位 us
	uint64_t tx = 0, rx =0;					//链路基于udp协议之上的数据收发总量，业务设置的统计间隔是2秒，则表示2秒内的数据总量，单位字节
	uint64_t last_tx_time = 0, last_rx_time = 0; //链路最后一次收发数据的时间，单位秒
	uint64_t create_time = 0;				//链路创建时间
	uint64_t lost = 0;						//链路丢包总数
	uint64_t tx_all = 0, rx_all = 0, lost_all = 0;

	//以下信息为链路调试信息
	uint64_t pacing_rate = 0;		 /* bit per second */
	uint64_t cwin = 0;				 //字节数
	uint64_t bw_estimate = 0; 		 /* bit per second */

public:
	int fill_to_packet(packet_t* p);
	int read_from_packet(packet_t* p, packet_t::ptr_unpacker upk);
	int read_from_map(const message_variant_t::map_t& args_map);
};

typedef std::shared_ptr<path_state_t> ptr_path_state_t;

//----------------------------------------------------------
//拥塞状态检查:
//在指定的时间窗口（单位毫秒，滑窗）内，利用线性回归方法统计
//和预测拥塞的趋势。
//每次调用:
//    update(时间/*毫秒*/, 当前待发缓冲区的大小/*字节*/)
//返回线性回归斜率:
//    “斜率”表示的是以1毫秒(ms)为单位的数据拥塞增长值(字节)。 
//    >0表示增长; =0表示平衡; <0表示下降
//----------------------------------------------------------

class congestion_detection {
protected:
	typedef std::pair<int64_t /*time*/, int64_t /*cache*/> rec_t;
	typedef std::queue<rec_t> sample_queue_t;

protected:
	int64_t		m_time_range = 1000;
	sample_queue_t	m_records;
	int64_t		m_first_time = 0;
	int64_t		m_sum_x = 0;
	int64_t		m_sum_y = 0;
	int64_t		m_sum_xx = 0;
	int64_t		m_sum_xy = 0;

	std::atomic<int64_t>	m_slope;
	std::atomic<int64_t>	m_duration;

protected:
	//first_time调整，重新计算所有的值
	void adjust_time_base();

public:
	congestion_detection(int32_t time_in_ms);
	~congestion_detection();

	void reset(int32_t time_in_ms = 0);
	void update(int64_t time_in_ms, int64_t cache_size);
	
	//atomic免锁操作
	inline double get_slope() { int64_t slope = m_slope; return (double)slope/10000.0; }
	inline int64_t get_duration() { return m_duration; }
};

typedef std::shared_ptr<congestion_detection>	ptr_congestion_detection_t;

//--------------------------------------------------------
//基本的KILOLINK连接上下文
//--------------------------------------------------------
class connection_context_t
{
friend class link_t;
public:
	connection_context_t();
	virtual ~connection_context_t();

//事件：派生类应该改写这些方法以完成正确的动作。
public:
	//创建初始化（握手/身份验证）包 - 通常在客户端模式下回调
	//（缺省）什么都不做，返回nullptr。表示这个连接不需要进行握手/身份验证操作。
	// 如果需要进行握手/身份验证，则应在方法中创建一个握手/身份验证packet，并将其返回。
	virtual ptr_packet_t on_create_initial_control_packet(klnk_stream_id_t stream_id) { return nullptr; };

	//当在CONTROL-STREAM上收到了一个包（意味着是CONTROL包），将回调。
	//注：当exception标志为true时，表示在接收数据过程中出现了异常。此时，packet可能是一个非完整的包，不可信。
	//返回一个有效的packet智能指针，表示这是要响应的消息packet；返回nullptr表示无需响应。
	//done_ret_flag 用于给调用者返回一个标识 - 告诉调用者这个包是否被理解和处理。这对于派生类需要调用父类的on_xxx方法比较有用。
	virtual ptr_packet_t on_received_control_packet(klnk_stream_id_t stream_id, ptr_packet_t packet, bool exception, int fin, bool& done_ret_flag) {return nullptr;};

	//在非CONTROL-STREAM上收到的包(stream packet)，将回调。
	//注：当exception标志为true时，表示在接收数据过程中出现了异常。此时，packet可能是一个非完整的包，不可信。
	//返回一个有效的packet智能指针，表示这是要响应的消息packet；返回nullptr表示无需响应。
	//done_ret_flag 用于给调用者返回一个标识 - 告诉调用者这个包是否被理解和处理。这对于派生类需要调用父类的on_xxx方法比较有用。
	virtual ptr_packet_t on_received_stream_packet(klnk_stream_id_t stream_id, ptr_packet_t packet, bool exception, int fin,  bool& done_ret_flag) {return nullptr;};

	//当底层KLNK的这个Connection已经断开，将回调on_destory表示应该释放该上下文。
	virtual void on_destroy() {};

public:
	//强制关闭当前上下文。注意：
	//1) 正常情况下，上下文是“被动”地被关闭（回调on_destroy事件）。上下文被关闭（on_destroy被回调）时，
	//   底层的QUIC连接已经被关闭/断开。所以，在on_destroy时，不需要考虑底层连接问题。
	//
	//2) force_close 用于强制关闭底层QUIC连接，这通常意味着是外界主动要断开QUIC连接。
	//
	//3) 线程安全。
	virtual void force_close();

	//在当前上下文发送一个包。
	virtual ssize_t send_packet(klnk_stream_id_t stream_id, ptr_packet_t packet, int fin = 0);
	virtual ssize_t send_packet(klnk_stream_id_t stream_id, packet_t* packet, int fin = 0);
	virtual ssize_t send_packet(klnk_stream_id_t stream_id, packet_t& packet, int fin = 0);

	//获取下一个可用的本地流ID。
	//is_unidir: 仅单向？默认为false表示双向流。
	klnk_stream_id_t get_next_local_stream_id(bool is_unidir = false);

	//获得当前上下文的CONTROL-STREAM id.
	inline klnk_stream_id_t get_control_stream_id() {return m_control_stream_id; }
	//对端的CONTROL-STREAM ID.
	inline klnk_stream_id_t get_peer_control_stream_id() {return m_peer_control_stream_id; }
	
	//当对方的第一个'KLNKCTL'包收到，表示链接确立。
	virtual bool link_ready();

	//从创建时起、到目前为止，这个上下文的持续时间。通常用来检查握手、身份验证的超时。
	//因为当context创建时，也是第一个握手/身份验证包的发出时间。通过检查 time_span() 是否
	//超过指定时间来确定握手/身份验证超时。
	//单位：us（1/1000000秒）
	/*virtual*/ int64_t time_span();

	//获得当前上下文的alpn
	const char* get_alpn(void) { return m_alpn.c_str(); }
	
	//获得当前上下文的连接模式（服务器/客户端模式）
	kilolink::link_mode_t get_link_mode(void) { return m_link_mode; }

	//（在CONTROL-STREAM上）发送控制包，这实际上是send_packet的变种。
	inline ssize_t send_control_packet(ptr_packet_t packet, int fin = 0) { return send_packet(get_control_stream_id(), packet, fin);}
	inline ssize_t send_control_packet(packet_t* packet, int fin = 0) { return send_packet(get_control_stream_id(), packet, fin);}
	inline ssize_t send_control_packet(packet_t& packet, int fin = 0) { return send_packet(get_control_stream_id(), packet, fin);}

	inline ssize_t send_control_ack_packet(ptr_packet_t packet, int fin = 0) { return send_packet(get_peer_control_stream_id(), packet, fin);}
	inline ssize_t send_control_ack_packet(packet_t* packet, int fin = 0) { return send_packet(get_peer_control_stream_id(), packet, fin);}
	inline ssize_t send_control_ack_packet(packet_t& packet, int fin = 0) { return send_packet(get_peer_control_stream_id(), packet, fin);}

	inline uint64_t get_data_sent() { return klnk_get_data_sent(m_connect_ctx); }
	inline uint64_t get_bytes_queued() { return klnk_get_bytes_queued(m_connect_ctx); }

	ptr_congestion_detection_t create_congestion_detection(int32_t time_range_ms);
	void remove_congestion_detection(ptr_congestion_detection_t cg);

protected:
	virtual void reset(klnk_instance_t inst, klnk_connection_t conn_ctx, kloop_t loop, uint64_t status_interval, const std::string& alpn = "*" );
	//用于内部的清理 - 它将在Loop线程上下文环境中，对对象进行内部清理，通常是在on_destroy之后
	virtual void internal_cleanup() {}

	//Status定时器，是connection context的“标配”功能。虽然在这个类中什么也不做。
	virtual void start_status_timer() {}
	virtual void stop_status_timer() {}

	//传输通道的数据统计
	virtual void on_transport_calculate(uint64_t send, uint64_t sent, uint64_t queue, uint64_t cache, uint64_t interval_time) {}

protected:
	void push_stream_data(klnk_stream_id_t stream_id, const uint8_t* buf, size_t buf_len, int fin);

protected:
	std::atomic_flag m_context_invalid = ATOMIC_FLAG_INIT;
	std::string		m_alpn;
	link_mode_t		m_link_mode = kilolink::mode_client;
	klnk_instance_t m_inst = nullptr;
	klnk_connection_t m_connect_ctx = nullptr;	
	kloop_t			m_context_loop = nullptr;
	uint64_t		m_status_interval = 0;

	//记录当前每一个流上正在接收（但尚未接收完整）的包。
	std::map<klnk_stream_id_t, ptr_packet_t> m_stream_pending_packets;

protected:
	klnk_stream_id_t m_control_stream_id = INVALID_STREAM_ID;
	klnk_stream_id_t m_peer_control_stream_id = INVALID_STREAM_ID;
	int64_t m_initial_time = 0;
	bool m_see_first_control_packet = false;

public:
	static void cg_timer_fn(kloop_t loop, kloop_timer_t timer_obj, void* userdata );
	void start_cg();
	void stop_cg();
	void update_cg();

protected:
	std::mutex m_cg_lock;
	kloop_timer_t m_cg_timer = nullptr;
	std::list<ptr_congestion_detection_t> m_cgs;
};

typedef std::shared_ptr<kilolink::connection_context_t> ptr_connection_context_t;

//--------------------------------------------------------
//缺省实现的 Connection context:
// 1) 带有缺省的身份验证行为（提供设备id和校验码）
// 2) 周期性地与对方交换路径状态信息和QUIC链路状态信息
//--------------------------------------------------------
class default_connection_context_t : public connection_context_t {
public:
	default_connection_context_t(	const std::string& default_auth_id = "", 
									const std::string& default_auth_code = "",
									const std::string& default_cli_type = "device",
									const std::string& default_cli_model = "UNKNOWN" );
	~default_connection_context_t();

public:
	//当身份验证时回调 - （需要派生类改写）
	//返回true表示鉴权成功，false表示失败。
	//返回时，填写 return_code 为返回代码；return_message 为返回消息。
	virtual bool on_authorize(	const std::string& auth_id,   //认证ID（设备ID）
								const std::string& auth_code, //授权码
								const std::string& cli_type,  //客户端类型（暂未明确定义，设备将暂时表示为device）
								const std::string& cli_model, //客户端（设备）型号
								int32_t& return_code, std::string& return_message )
	{ return true; } //TODO: 派生类必须改写，否则不做身份验证。

	//（客户端）当身份认证请求被远程服务器应答后回调。
	virtual void on_authorize_acked(int32_t ret_code, const std::string& ret_msg) {}

	//收到了远程更新path-state的请求后回调（通知）
	virtual void on_peer_path_updated(bool valid) {}

public: //重载
	ptr_packet_t on_create_initial_control_packet(klnk_stream_id_t stream_id);
	ptr_packet_t on_received_control_packet(klnk_stream_id_t stream_id, ptr_packet_t packet, bool exception, int fin, bool& done_ret_flag);
	bool link_ready(); 

public:
	virtual bool is_authorized(void) { return m_authorized; }

	//获取对端的路径状态
	inline int get_peer_path_states(std::map<std::string, ptr_path_state_t>& ret_map) {
		ret_map.clear();
		std::unique_lock<std::recursive_mutex> lck(m_peer_path_lock);
		ret_map.insert( m_peer_path_states.begin(), m_peer_path_states.end() );
		return (int)ret_map.size();
	}

	//获取本端的路径状态
	inline int get_local_path_states(std::map<std::string, ptr_path_state_t>& ret_map) {
		ret_map.clear();
		std::unique_lock<std::recursive_mutex> lck(m_local_path_lock);
		ret_map.insert( m_local_path_states.begin(), m_local_path_states.end() );
		return (int)ret_map.size();
	}

protected:
	static void timer_cb(kloop_t loop, kloop_timer_t timer_obj, void* userdata );
	void update_local_path_info();

protected: //overwrite.
	void reset(klnk_instance_t inst, klnk_connection_t conn_ctx, kloop_t loop, uint64_t status_interval, const std::string& alpn = "*" );
	void internal_cleanup();
	void start_status_timer();
	void stop_status_timer();

protected:
	std::string			m_default_auth_id = nullptr;
	std::string			m_default_auth_code = nullptr;
	std::string			m_default_cli_type = nullptr;
	std::string			m_default_cli_model = nullptr;

	//timer.
	kloop_timer_t 		m_status_timer = nullptr;

	bool				m_authorized = false;
	klnk_stream_id_t	m_pathinfo_stream_id = INVALID_STREAM_ID;
	klnk_stream_id_t	m_peer_pathinfo_stream_id = INVALID_STREAM_ID;

	std::recursive_mutex	m_local_path_lock;
	std::map<std::string, ptr_path_state_t>	m_local_path_states;

	std::recursive_mutex	m_peer_path_lock;
	std::map<std::string, ptr_path_state_t>	m_peer_path_states;
};

//--------------------------------------------------------

//用于创建connection上下文的工厂方法。
typedef std::function<ptr_connection_context_t(klnk_instance_t/*inst*/, klnk_connection_t/*connect_ctx*/, const std::string& /*alpn*/)> connection_context_factory_fn;

//路径创建时的自定义方法
//typedef std::function<> path_create_fn;

//--------------------------------------------------------
//LINK type
//--------------------------------------------------------
class link_t
{
public:
	link_t();
	~link_t();

public:
	int create(const options_t &options, kloop_t loop = nullptr);
	void destroy();

	//注册 - 依据ALPN创建连接上下文(connection_context)的工厂方法
	bool register_connection_context_factory(const std::string& alpn, connection_context_factory_fn fn);
	inline bool unregister_connection_context_factory(const std::string& alpn) { return register_connection_context_factory(alpn, nullptr); }

	//注册 - 默认创建上下文的工厂方法
	bool register_connection_context_factory(connection_context_factory_fn fn);
	inline bool unregister_connection_context_factory(void) { return register_connection_context_factory(nullptr); }

public:
	int path_add(const char *path_if);
	int path_del(const char *path_if);

protected:
	//Lower-level KLNK回调事件
	static void klnk_connection_established(klnk_instance_t inst, const klnk_connection_id_t *connect_id, klnk_connection_t connect_ctx);
	static void klnk_connection_disconnected(klnk_instance_t inst, const klnk_connection_id_t *connect_id, klnk_connection_t connect_ctx);
	static void klnk_loop_timer_send_path_info(kloop_t loop, kloop_timer_t timer_obj, void *userdata);
	static void klnk_recv_data(klnk_instance_t inst, klnk_connection_t connect_ctx, uint64_t stream_id, int is_fin, const uint8_t *buf, size_t buf_size, int *change);
	static void klnk_instance_loop(void *arg);
	static void klnk_transport_calculate(klnk_instance_t inst, klnk_connection_t connect_ctx, uint64_t send, uint64_t sent, uint64_t queue, uint64_t cache, uint64_t interval_time);

protected:
	//（connection_ctx <-> 网络上下文）为映射的记录
	std::map<klnk_connection_t, ptr_connection_context_t>	m_connections;
	std::map<std::string, connection_context_factory_fn>	m_context_factories;
	std::mutex m_connections_lock;

protected:
	options_t m_options;
	std::thread m_run_task;
	klnk_socket_t m_socket = -1;

	//klnk lower level.
	klnk_create_options_t m_opt;
	klnk_instance_t m_inst = nullptr;
};

}

#endif