#ifndef __KLNK_ADVANCED_H__
#define __KLNK_ADVANCED_H__

#include "klnk.hpp"
#include <atomic>
#include <thread>
#include <queue>

namespace kilolink {

//KiloLink高级版的连接上下文实现。
//支持动态创建的VPN, TCP端口转发和UDP端口转发
class advanced_connection_context_t : public default_connection_context_t {
protected:
	//-------------------------------------------------
	enum operator_flag {
		flag_none = 0,
		flag_adding = 1,
		flag_removing = 2,
		flag_init = 3,
		flag_pending = 4, //正在解析目标地址
		flag_dead = 5
	};

	enum forward_type {
		fwd_unset = 0,
		fwd_tcp = 1,
		fwd_udp = 2,
		fwd_vpn = 3
	};

	//-------------------------------------------------
	struct forward_detail {
		std::atomic<int>			op_flag;
		forward_type				fwd_type = fwd_unset;
		int 						fd = -1;
		int							own_fd = -1;
		klnk_stream_id_t			stream_id = INVALID_STREAM_ID;
		bool						no_managed = false;

		std::string 				target;
		uint16_t					target_port = 0;
		struct sockaddr_storage		target_addr_sto;

		uint64_t					last_tx_rx_time;
		std::thread					backend_task;

		std::mutex					lock;
		kloop_io_t					rx_io = nullptr, tx_io = nullptr;
		bool						rx_on = false, tx_on = false;
		std::queue<ptr_packet_t>	pending_pkts;
	};

	//-------------------------------------------------
	typedef std::shared_ptr<forward_detail>	ptr_forward_detail;

public:
	//指定一个run_loop - 注意它与link_t的loop不是指同一概念。
	//link_t的loop用于驱动QUIC的工作，但这个上下文的run_loop用于驱动进行
	//端口转发时的套接字IO. 它们可以是同一个，但也可以不是同一个loop.
	advanced_connection_context_t(	kloop_t run_loop = nullptr,
									const std::string& default_auth_id = "", 
									const std::string& default_auth_code = "",
									const std::string& default_cli_type = "device",
									const std::string& default_cli_model = "UNKNOWN" );

	~advanced_connection_context_t();

	void set_default_tuntap_device(const std::string& device_name = "/dev/net/tun");

protected:
	static void on_period_timer(kloop_t loop, kloop_timer_t timer_obj, void* userdata );
	static void on_update_async(kloop_t loop, kloop_async_t async_obj, void* userdata );
	static void on_ioevt_async(kloop_t loop, kloop_async_t async_obj, void* userdata );
	static void on_tcp_listen_io_read(kloop_t loop, kloop_io_t io_obj, int fd, void* userdata, kloop_io_event_t event );
	static void on_tcp_connection_io_read(kloop_t loop, kloop_io_t io_obj, int fd, void* userdata, kloop_io_event_t event );
	static void on_tcp_connection_io_write(kloop_t loop, kloop_io_t io_obj, int fd, void* userdata, kloop_io_event_t event );
	static void on_udp_listen_io_read(kloop_t loop, kloop_io_t io_obj, int fd, void* userdata, kloop_io_event_t event );
	static void on_udp_connection_io_read(kloop_t loop, kloop_io_t io_obj, int fd, void* userdata, kloop_io_event_t event );
	static void on_vpn_connection_io_read(kloop_t loop, kloop_io_t io_obj, int fd, void* userdata, kloop_io_event_t event );
	static void on_vpn_connection_io_write(kloop_t loop, kloop_io_t io_obj, int fd, void* userdata, kloop_io_event_t event );
	
	//注：线程参数为copy类型的字符串，以防止对象析构时的内存访问异常
	void task_thread_run(std::string auth_id);
	void period_check();
	void update_forwarding();
	void check_and_start_txs();

	//-----------------------------------------
	void start_rx(ptr_forward_detail detail);
	void stop_rx(ptr_forward_detail detail);
	void start_tx(ptr_forward_detail detail);
	void stop_tx(ptr_forward_detail detail);

	bool create_socket_listen_instance(ptr_forward_detail detail);
	bool cleanup_socket_listen_instance(ptr_forward_detail detail);

	bool create_socket_connection_instance(ptr_forward_detail detail, bool is_accepted);
	bool cleanup_socket_connection_instance(ptr_forward_detail detail);

	//-----------------------------------------
	void process_tcp_listen_io_read(int listen_fd);
	void process_tcp_connection_io_read(int fd);
	void process_tcp_connection_io_write(int fd);
	void process_udp_listen_io_read(int listen_fd);
	void process_udp_connection_io_read(int fd);

	void process_vpn_connection_io_read(int fd);
	void process_vpn_connection_io_write(int fd);
	//-----------------------------------------
	void dynamic_connect_thread_task(ptr_forward_detail detail);

public: //静态方法
	static int open_tcp_listen_socket(uint16_t *listen_port);
	static int open_tcp_listen_socket(const std::string& bind_addr, uint16_t *listen_port);

	static int open_tcp_connect_socket(const std::string& to_addr, uint16_t to_port);
	static int open_tcp_connect_socket(struct sockaddr_storage* to_addr);

	static int open_udp_socket(uint16_t *listen_port);
	static int open_udp_socket(const std::string& bind_addr, uint16_t *listen_port);

	//注意: ifname如果为空，打开成功后，将返回实际创建的ifname
	static int open_tuntap_vpn(const std::string& tuntap_dev, std::string& ifname, const std::string& mode = "tun");

	//对指定的网络接口（主要考虑针对vpn接口）设定网络选项。
	//options: 以";"为分割的key1:value1; key2:value 2; ... 形式。
	//已定义以下选项：
	//	mode: tun/tap
	//	ifname: 强行指定的接口名称
	//	hwaddr: 硬件(物理MAC)地址
	//	ip: IPv4地址/或"dhcp"
	//	ip6: IPv6地址/或"dhcp"
	//	prefix: 掩码(16/24/...)
	//	pointopoint: 点对点对端地址
	//	gw: 默认网关地址
	//	mtu: MTU大小
	//
	//	TODO: 以下命令暂时还没有实现。
	//	active_cmd: 在网卡创建时要执行的命令(例如route)
	//  active_cmd: ... (可以有多个active命令)
	//	deactive_cmd: 在网卡销毁时要执行的命令(例如route)
	//  deactive_cmd: ... (可以有多个active命令)
	//
	//	注意：在命令中，如果涉及到接口名称的部分，可以用 %ifname% 代替。
	static inline bool set_if_options(const std::string& ifname, const std::string& options) {
		std::multimap<std::string, std::string> options_map;
		parse_if_options(options, options_map);
		return set_if_options(ifname, options_map);
	}
	
	//将选项解析成map后设定。
	static bool set_if_options(const std::string& ifname, const std::multimap<std::string, std::string>& options);

	//将字符串形式的选项解析成map.
	static int parse_if_options(const std::string& options, std::multimap<std::string, std::string>& options_map);

public:
	//注意：以下 no_managed 标志用于表示该fd不被内部监管，这意味着在关闭context时，不会close(fd)
	//      这适用于由你的程序管理fd，当每次connection context动态创建和销毁时，不会造成socket的关闭/再打开。
	int add_tcp_forward( int sock_fd, const std::string& target_addr, uint16_t target_port, bool no_managed = false );
	int add_udp_forward( int sock_fd, const std::string& target_addr, uint16_t target_port, bool no_managed = false );

	//options参见set_if_options
	int add_vpn_forward( int tun_tap_fd, const std::string& options, bool no_managed = false );

	//移除已添加的转发任务。
	//handle: add_tcp_forward/add_udp_forward/add_vpn_forward所返回的值。
	bool remove_forward(int handle);

public: //回调事件
	//当远端发起一个新的TCP转发请求时回调:
	//	stream_id = 远端发起通过的stream_id
	//	to_addr = 请求连接的目标地址（已经过内部解析后的）
	//返回：
	//	如果由你自己创建了一个TCP socket，请返回创建的socket fd; 否则返回 <0 的值交由内部自动创建。
	virtual int on_create_tcp_connect_socket( klnk_stream_id_t stream_id, struct sockaddr_storage* to_addr);

	//当远端发起一个新的UDP转发请求时回调:
	//	stream_id = 远端发起通过的stream_id
	//	to_addr = 请求连接的目标地址（已经过内部解析后的）
	//返回：
	//	如果由你自己创建了一个UDP socket，请返回创建的socket fd; 否则返回 <0 的值交由内部自动创建。
	virtual int on_create_udp_connect_socket( klnk_stream_id_t stream_id, struct sockaddr_storage* to_addr);

	//当远端发起一个新的VPN转发请求时回调:
	//	stream_id = 远端发起通过的stream_id
	//	options_map = 经过解析后的VPN选项参数map (multimap)，详见set_if_options说明
	//	ifname = 如果由你自己创建了VPN，请返回实际创建的vpn的名称
	//	mode = VPN模式: tun/tap
	//返回：
	//	如果由你自己创建了一个VPN，请返回创建VPN后的接口fd; 否则返回 <0 的值交由内部自动创建。
	virtual int on_create_vpn_interface( klnk_stream_id_t stream_id, std::multimap<std::string, std::string>& options_map, std::string& ifname, const std::string& mode );

	//当数据包不是FWDR的时候需要外部处理
	//	stream_id  = 远端发起通过的stream_id
	//	packet = 数据包
	//	exception = 
	//	fin = 当前stream是否结束
	//	done_ret_flag = true表示被处理，false表示没有被处理
	//返回：
	//	表示是否有数据包需要返回
	virtual ptr_packet_t on_external_analyse_stream_packet(klnk_stream_id_t stream_id, ptr_packet_t packet, bool exception, int fin,  bool& done_ret_flag);

public: //overwrite
	void reset(klnk_instance_t inst, klnk_connection_t conn_ctx, kloop_t loop, uint64_t status_interval, const std::string& alpn = "*" );
	ptr_packet_t on_received_stream_packet(klnk_stream_id_t stream_id, ptr_packet_t packet, bool exception, int fin,  bool& done_ret_flag);
	void on_destroy();

protected:
	std::string								m_tun_tap_device = "/dev/net/tun";
	kloop_t									m_run_loop = nullptr;
	kloop_timer_t							m_period_timer = nullptr;
	kloop_async_t							m_update_async = nullptr;
	kloop_async_t							m_ioevt_async = nullptr;
	std::atomic<bool>						m_terminate;
	std::thread								m_run_task;

	klnk_stream_id_t						m_last_seen_sid = 0;

	std::mutex								m_fwd_lock;
	std::mutex								m_ioevt_lock;

	std::queue<ptr_forward_detail>			m_tx_ioevt;

	//Forwarding有三种场景：
	//1. Listen(er) - 监听指定的端口（或VPN）
	std::map<int, ptr_forward_detail>				m_listen;

	//2. 因Listener而接收到的本地连接实例
	std::map<int, ptr_forward_detail>				m_accepted;
	//特例情况 - UDP可能存在多个不同地址到listener端口的数据，需要依地址进行区分并假设为一个一个独立的实例
	std::map<std::string, ptr_forward_detail>		m_udp_accepted;
	std::map<klnk_stream_id_t, ptr_forward_detail>	m_accepted_by_sid;

	//3. 远端请求过来的（需要由我中继转出的）连接，属于被动状态下的socket或vpn接口
	std::map<int, ptr_forward_detail>				m_connected;
	std::map<klnk_stream_id_t, ptr_forward_detail>	m_connected_by_sid;

	//4. 正在等候确认连接的。
	std::map<klnk_stream_id_t, ptr_forward_detail>	m_connect_pending;
};

}

#endif
