-----------------------------------------------------------------------------------------------------
-- FPGA_Rx chip controller.
-- Copyright (C) 2016, KILOVIEW Elecontrics Co., Ltd.
--
-- Author: 	Jacob Zuo, 2019/3/18
-----------------------------------------------------------------------------------------------------
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local periphery = require "periphery"
local hdtv_formats = require "chips.hdtv_formats"
local I2C_REG = require "chips.i2c_reg"

require "ndi_encoding"

local eReg_Video 	= 0x200
local eReg_Audio 	= 0x220

local locked 		= 0x01
local Mode_12G		= 0x05
local Mode_6G		= 0x04
local Mode_3G		= 0x02
local Mode_HD		= 0x00
local Mode_SD		= 0x01

local Rate_60		= 0x0b
local Rate_59_94	= 0x0a
local Rate_50		= 0x09
local Rate_48 		= 0x08
local Rate_30		= 0x07
local Rate_29_97 	= 0x06
local Rate_25		= 0x05
local Rate_47_95    = 0x04
local Rate_24		= 0x03
local Rate_23_98 	= 0x02

local vid_1920x1080p0	= 0x00
local vid_1920x1080p1	= 0x03
local vid_2048x1080p	= 0x02

local vid_720p			= 0x01
local vid_720x486		= 0x08
local vid_720x576		= 0x09

-- sdi_t_family
--0000 SMPTE ST 274 [Ref 8] 1920 x 1080
--0001 SMPTE ST 296 [Ref 9] 1280 x 720
--0010 SMPTE 2048-2 [Ref 10] 2048 x 1080
--0011 SMPTE 295 [Ref 11] 1920 x 1080
--1000 NTSC 720 x 486
--1001 PAL 720 x 576
--1111 Unknown
--Others Reserved
------------------------------------------------------
-- The factory function to create a FPGA_Rx handle object
------------------------------------------------------
local function CreateFPGA_Rx(i2c_dev, rxAddr)
	------------------------------------------------------
	-- Object
	------------------------------------------------------
	local FPGA_Rx = {
		currentSource = "SDI",
		currentSignal = false,
		freeRun = false,
		videoFormat = { _valid = false ,_formatInfo={}},
		audioFormat = { _valid = false, audio_src = "SDI", audio_signal_type = 0, channels = 0 }, --type 0 aka. SDI
		orgAudioFormat = { withAudio = false, channels = 0 },
		nostandardFormatInfo = {},
		sdi_infor = {},
		index = 0
	}

	i2c_dev = i2c_dev or "/dev/i2c-1"
	
	local BASE_MAP = rxAddr or 0x69

	_dbg.Info("FPGA_Rx", "Creating FPGA_Rx control object at 0x" .. string.format( "%02x", BASE_MAP) .. "@" .. i2c_dev .. "..." )

	FPGA_Rx.i2c_reg = I2C_REG.Create()
	if not FPGA_Rx.i2c_reg:Open( i2c_dev ) then
		return nil, "Fail to open I2C device"
	end

	-------------------------------------------------------
	-- Object method: Register read
	-------------------------------------------------------
	function FPGA_Rx:ReadReg( map, addr, itemArr )
		return self.i2c_reg:ReadReg(map, addr, itemArr)
	end

	------------------------------------------------------
	-- Register write
	------------------------------------------------------
	function FPGA_Rx:WriteReg( map, addr, itemArr )
		return self.i2c_reg:WriteReg(map, addr, itemArr)
	end

	------------------------------------------------------
	-- Soft reset the chip!
	------------------------------------------------------
	function FPGA_Rx:Reset()
		self:_InitDevice()

		self.currentSignal = false
		self.freeRun = false
		self.audioFormat = { _valid = false, audio_src = "SDI", audio_signal_type = 0, channels = 0 }
		self.orgAudioFormat = { withAudio = false, channels = 0 }
		self.videoFormat = {_valid = false,_formatInfo={}}
		self.sdi_infor = {
			f_12g = 0x10,
			sdi_locked = 0x00,
			progressive = 0x00,
			levelb	= 0x00,
			sdi_mode = 0x00,
			sdi_family = 0x00,
			sdi_rate = 0x00,
			sdiformat = 0x0000
		}
	end

	------------------------------------------------------
	-- Init the FPGA_Rx Device modify ok
	------------------------------------------------------
	function FPGA_Rx:_InitDevice()
		self:WriteReg( BASE_MAP,0xA2,{0x00} ) -- 有些板子切换编解码模式会让fpga复位，有些不会？
		--TODO: Initialize device.
	end

	------------------------------------------------------
	-- Init the FPGA_Rx interrupts
	------------------------------------------------------

	function FPGA_Rx:_InitInterrupts()
		--TODO: Nothing yet.
	end

	------------------------------------------------------
	--check the input signal is valid
	------------------------------------------------------
	function FPGA_Rx:_CheckSignal()
		--local freeRun = false
		--TODO: check signal.

-- 	    0XA1: R  sdi 采集信息
--      bit7_7:  0:sdi无输入  1:sdi有输入
--      bit6_6: 0:整数频率，如 60，30等 1:小数频率，如 59.94，29.97
--      bit5_4: 分辨率(0:SD, 1:1280x720, 2:1920x1035, 3:1920x1080)
--      bit3_1: 帧率(0,1,,7 ---> res,24p,25p,30p,50i,60i,50p,60p)
-- 	 0xA7:  R video_transport[7:0]
--   bit7_2 : 保留
--   bit1_0 : i=00, psf=01, p=11
		local sdif = self:ReadReg(BASE_MAP,0xA1)
		local sdiformat = sdif and sdif[1] or 0x36

		local bit7_7 = (sdiformat&0x80)>>7
		local bit6_6 = (sdiformat&0x40)>>6
		local bit5_4 = (sdiformat&0x30)>>4
		local bit3_1 = (sdiformat&0x0e)>>1


		local transport = self:ReadReg(BASE_MAP,0xa7)
		local video_transport = transport and  transport[1]&0x03 or 0


		self.sdi_infor.f_12g = 0x10

		local locked = ((sdiformat&0x80)>>7)
		if locked == 1 and self.index<5 then
			self.index = self.index + 1
		elseif locked ==0 then
			self.index = 0
		end
		self.sdi_infor.sdi_locked = self.index>4 and 1 or 0
		self.sdi_infor.progressive = 0x01 -- 0:i,1:p,2:Psf
		self.sdi_infor.levelb = 0x00
		self.sdi_infor.sdi_mode = Mode_HD
		self.sdi_infor.sdi_family = vid_1920x1080p0
		self.sdi_infor.sdi_rate = 0x00


		if bit3_1==4 or bit3_1==5 or (bit3_1==1 and video_transport==1) then
			self.sdi_infor.progressive = 0
		end

		if bit5_4==0 then
			self.sdi_infor.sdi_mode = Mode_SD
		elseif bit5_4==3 and (bit3_1==6 or bit3_1==7) then
			self.sdi_infor.sdi_mode = Mode_3G
		end

		if bit5_4==1 then
			self.sdi_infor.sdi_family = vid_720p
		elseif bit5_4 == 0 and bit3_1==4 then
			self.sdi_infor.sdi_family = vid_720x576
		elseif  bit5_4 == 0 and bit3_1==5 then
			self.sdi_infor.sdi_family = vid_720x486
		end

		if bit6_6==1 then -- 小数帧率
			if bit3_1==1 then self.sdi_infor.sdi_rate=0x02
			elseif bit3_1==3 then self.sdi_infor.sdi_rate=0x06
			elseif bit3_1==5 then self.sdi_infor.sdi_rate=0x06
			elseif bit3_1==7 then self.sdi_infor.sdi_rate=0x0a
			end
		else -- 整数帧率
			if bit3_1==1 then self.sdi_infor.sdi_rate=0x03
			elseif bit3_1==2 then self.sdi_infor.sdi_rate=0x05
			elseif bit3_1==3 then self.sdi_infor.sdi_rate=0x07
			elseif bit3_1==4 then self.sdi_infor.sdi_rate=0x05
			elseif bit3_1==5 then self.sdi_infor.sdi_rate=0x07
			elseif bit3_1==6 then self.sdi_infor.sdi_rate=0x09
			elseif bit3_1==7 then self.sdi_infor.sdi_rate=0x0b
			end
		end


		--[[
		_dbg.Info( "CODEC", "sdi_information:", self.sdi_infor.f_12g,
			self.sdi_infor.sdi_locked,
			self.sdi_infor.progressive,
			self.sdi_infor.levelb,
			self.sdi_infor.sdi_mode,
			self.sdi_infor.sdi_family,
			self.sdi_infor.sdi_rate
		)
		--]]

		self.sdi_infor.sdiformat = sdiformat

		return self.sdi_infor.sdi_locked == 0x01
	end

	------------------------------------------------------
	-- A step handler for FPGA_Rx processing.
	-- To fit the LUA working mode, you need call the 
	-- 'Step' periodically ( suggest per 200~500ms).
	-- This routine should do :
	--	* Signal status checking
	--	*	Auto format detecting
	--	*	Configue the chip to make it output BT.656/BT.601 signals
	--
	------------------------------------------------------
	
	function FPGA_Rx:Step()
		self.currentSignal, self.freeRun = self:_CheckSignal()	--Check current signal

		--TODO: consider free-run but now not support.
		if self.currentSignal ~= self.videoFormat.signal_locked then
			if not self.videoFormat.version then --Keep track the updated version
				self.videoFormat.version = 0
			else
				self.videoFormat.version = self.videoFormat.version + 1
			end
		end

		if not self.audioFormat.version then
			self.audioFormat.version = 0
		end

		if self.currentSignal then
			self.videoFormat.signal_locked = true	--Assume it be, but when detecting, it may changed
			self.videoFormat.free_run = false
			self:_DetectCurrentFormat()
		elseif self.freeRun then
			self.videoFormat.signal_locked = true
			self.videoFormat.free_run = true
		else
			self.videoFormat.signal_locked = false
			self.videoFormat.free_run = false
		end
		
	--	if self.videoFormat.signal_locked == true then
	--		print("locked")
	--	else
	--		print("unlocked")
	--	end

	end


	------------------------------------------------------
	-- Detect current signal format
	------------------------------------------------------
	function FPGA_Rx:_DetectCurrentFormat()
		self:_DetectSDIAudio()
		self:_UpdateAudioFormatInfo()
		self:_DetectSDIFormat()
	end

	------------------------------------------------------
	-- Detect SDI Audio
	------------------------------------------------------
	function FPGA_Rx:_DetectSDIAudio()
		local aud_chn_low_8 = self:ReadReg(BASE_MAP,0xa6)
		local aud_chn_high_8 = self:ReadReg(BASE_MAP,0xa5)
		local audio_pattern 
		if aud_chn_low_8 and aud_chn_high_8 then
			audio_pattern = (aud_chn_high_8[1]<<8)+aud_chn_low_8[1] -- 1 aka. Pattern.
		end
		if audio_pattern then
			local ch = 0
			for i=0,15 do
				if (audio_pattern & (1<<i)) ~= 0 then
					ch = ch + 1
				end
			end

			self.orgAudioFormat.sample = 48000 --TODO.
			self.orgAudioFormat.withAudio = (ch > 0)
			self.orgAudioFormat.channels = ch
			self.orgAudioFormat.fs = 256
		end

		return true
	end

	function FPGA_Rx:_UpdateAudioFormatInfo()
		if self.audioFormat._valid ~= self.orgAudioFormat.withAudio or
			self.audioFormat.sampling ~= self.orgAudioFormat.sample or
			self.audioFormat.channels ~= self.orgAudioFormat.channels then

			self.audioFormat.version = self.audioFormat.version + 1
			self.audioFormat._valid = self.orgAudioFormat.withAudio
			self.audioFormat.fs = self.orgAudioFormat.fs
			self.audioFormat.sampling = self.orgAudioFormat.sample
			self.audioFormat.channels = self.orgAudioFormat.channels
			self.audioFormat.bit_width = 16
		end
	end

	------------------------------------------------------
	-- According detected format information, record it to format context.
	------------------------------------------------------
	
	function FPGA_Rx:_RecordVideoFormatInfo(src, formatInfo)
		if type(formatInfo) ~= "table" then
			return false
		end
		--pretty.dump( formatInfo )
		
		if self.videoFormat._formatInfo ~= formatInfo or (src and src ~= self.videoFormat.video_src) then
			if not self.videoFormat.version then --Keep track the updated version
				self.videoFormat.version = 0
			else
				self.videoFormat.version = self.videoFormat.version + 1
			end

			--This flag is no used in current driver.
			--self.videoFormat._lastFormatChanged = true

			self.videoFormat._formatInfo = formatInfo
			
			if src then
				self.videoFormat.video_src = src
			end
			
			self.videoFormat.format_code = formatInfo.formatId	-- Make the formatId as format-code.
			self.videoFormat.format_name = formatInfo.formatName or (formatInfo.hAct .. "x" .. formatInfo.vAct0)
			if formatInfo.rateMode == hdtv_formats.VIRATE_4K then
				self.videoFormat.video_mode = "4k"
			elseif formatInfo.rateMode == hdtv_formats.VIRATE_4K60 then
				self.videoFormat.video_mode = "4k60"
			elseif formatInfo.rateMode == hdtv_formats.VIRATE_SD then
				self.videoFormat.video_mode = "sd"
			else
				self.videoFormat.video_mode = "hd"
			end

			return true
		else
			return true	--No changed.
		end
	end

	------------------------------------------------------
	function FPGA_Rx:SetVideoMode(sig,vidMod,colorSpace)
	end
	
	------------------------------------------------------
	-- Detect SDI input format
	------------------------------------------------------

	function FPGA_Rx:_DetectSDIFormat()
		--local signal_locked	= false
		local video_mode 		= hdtv_formats.VIRATE_HD
		local hAct				= 1920
		local vAct0				= 1080
		local vAct1				= 0
		local hTotal			= 2200
		local vTotal0 			= 1125
		local vTotal1			= 0
		local frameRate			= 30
		local rate_n			= 30000
		local rate_d  			= 1000
		local pxclk				= 74250000
		local format_name 		= "INVALID"
		local interlaced		= false

		if self.sdi_infor.sdi_locked == locked then
			--signal_locked = true
			if self.sdi_infor.progressive == 0x01 or self.sdi_infor.levelb == 0x01 then
				interlaced = false
			else
				interlaced = true
			end

			if self.sdi_infor.sdi_mode == Mode_12G then
				video_mode 		= hdtv_formats.VIRATE_4K60
				pxclk	= 594000000
				if self.sdi_infor.sdi_family == vid_2048x1080p then
					hAct	= 4096
					vAct0	= 2160
					hTotal	= 4400
					vTotal0 = 1125
				else
					hAct	= 3840
					vAct0	= 2160
					hTotal	= 4400
					vTotal0 = 2250
				end
				--print("Mode_12G")
			elseif self.sdi_infor.sdi_mode == Mode_6G then
				video_mode 		= hdtv_formats.VIRATE_4K
				pxclk	= 297000000
				if self.sdi_infor.sdi_family == vid_2048x1080p then
					hAct	= 4096
					vAct0	= 2160
					hTotal	= 4400
					vTotal0 = 1125
				else
					hAct	= 3840
					vAct0	= 2160
					hTotal	= 4400
					vTotal0 = 2250
				end
				--print("Mode_6G")
			elseif self.sdi_infor.sdi_mode == Mode_3G then
				video_mode  = hdtv_formats.VIRATE_3G
				pxclk	    = 148500000
				if self.sdi_infor.sdi_family == vid_2048x1080p then
					hAct	= 2048
					vAct0	= 1080
					hTotal	= 2200
					vTotal0 = 1125
				else
					hAct	= 1920
					vAct0	= 1080
					hTotal	= 2200
					vTotal0 = 1125
				end
				--print("Mode_3G")
			elseif self.sdi_infor.sdi_mode == Mode_HD then
				video_mode 		= hdtv_formats.VIRATE_HD
				pxclk	= 74250000
				if self.sdi_infor.progressive == 0x00 then
				-- Interlaced format: progressive == 0x00
					hAct	= 1920
					vAct0	= 540
					vAct1	= 540

					hTotal	= 2200
					vTotal0 = 562
					vTotal1 = 563

				elseif self.sdi_infor.sdi_family == vid_1920x1080p0 or self.sdi_infor.sdi_family == vid_1920x1080p1 then
					hAct	= 1920
					vAct0	= 1080
					hTotal	= 2200
					vTotal0 = 1125
				elseif self.sdi_infor.sdi_family == vid_2048x1080p then
					hAct	= 2048
					vAct0	= 1080
					hTotal	= 2200
					vTotal0 = 1125
				else			
					hAct	= 1280
					vAct0	= 720
					hTotal	= 1650
					vTotal0 = 750
				end
				--print("Mode_HD")
			elseif self.sdi_infor.sdi_mode == Mode_SD then --?????
				video_mode 		= hdtv_formats.VIRATE_SD
				pxclk	= 27000000
				if self.sdi_infor.sdi_family == vid_720x486 then
					--hAct	= 720
					--vAct0	= 243
					--hTotal	= 858
					--vTotal0 = 576

					hAct	= 720
					vAct0	= 240
					vAct1	= 240

					hTotal	= 858
					vTotal0 = 262
					vTotal1 = 263

				else 
					--hAct	= 720
					--vAct0	= 288
					--hTotal	= 858
					--vTotal0 = 576
					hAct	= 720
					vAct0	= 288
					vAct1	= 288

					hTotal	= 864
					vTotal0 = 312
					vTotal1 = 313			
				end
				--print("Mode_SD")
			end

			if self.sdi_infor.levelb == 0x01 then
				if self.sdi_infor.sdi_rate == Rate_30 then
					frameRate	= 60
					rate_n		= 60000
					rate_d  	= 1000
					--print("Rate_60")
				elseif self.sdi_infor.sdi_rate == Rate_29_97 then
					frameRate	= 59.94
					rate_n		= 60000
					rate_d  	= 1001
					--print("Rate_59_94")
				elseif self.sdi_infor.sdi_rate == Rate_25 then
					frameRate	= 50
					rate_n		= 50000
					rate_d  	= 1000
					--print("Rate_50")
				elseif self.sdi_infor.sdi_rate == Rate_24 then
					frameRate	= 48
					rate_n		= 48000
					rate_d  	= 1000
					--print("Rate_24")
				elseif self.sdi_infor.sdi_rate == Rate_23_98 then
					frameRate	= 47.95
					rate_n		= 47950
					rate_d  	= 1000
					--print("Rate_23_98")
				else
					frameRate	= 60
					rate_n		= 60000
					rate_d  	= 1000
					--print("Rate_60")
				end
				--print(self.sdi_infor.sdi_rate)
			else
				if self.sdi_infor.sdi_rate == Rate_60 then
					frameRate	= 60
					rate_n		= 60000
					rate_d  	= 1000
					--print("Rate_60")
				elseif self.sdi_infor.sdi_rate == Rate_59_94 then
					frameRate	= 59.94
					rate_n		= 60000
					rate_d  	= 1001
					--print("Rate_59_94")
				elseif self.sdi_infor.sdi_rate == Rate_50 then
					frameRate	= 50
					rate_n		= 50000
					rate_d  	= 1000
					--print("Rate_50")
				elseif self.sdi_infor.sdi_rate == Rate_48 then
					frameRate	= 48
					rate_n		= 48000
					rate_d  	= 1000
					--print("Rate_48")
				elseif self.sdi_infor.sdi_rate == Rate_30 then
					frameRate	= 30
					rate_n		= 30000
					rate_d  	= 1000
					--print("Rate_30")
				elseif self.sdi_infor.sdi_rate == Rate_29_97 then
					frameRate	= 29.97
					rate_n		= 30000
					rate_d  	= 1001
					--print("Rate_29_97")
				elseif self.sdi_infor.sdi_rate == Rate_25 then
					frameRate	= 25
					rate_n		= 25000
					rate_d  	= 1000
					--print("Rate_25")
				elseif self.sdi_infor.sdi_rate == Rate_47_95 then
					frameRate	= 47.95
					rate_n		= 47950
					rate_d  	= 1000
					--print("Rate_47_95")
				elseif self.sdi_infor.sdi_rate == Rate_24 then
					frameRate	= 24
					rate_n		= 24000
					rate_d  	= 1000
					--print("Rate_24")
				elseif self.sdi_infor.sdi_rate == Rate_23_98 then
					frameRate	= 23.98
					rate_n		= 24000
					rate_d  	= 1001
					--print("Rate_23_98")
				else
					frameRate	= 60
					rate_n		= 60000
					rate_d  	= 1000
					--print("else")
				end
			end

			if self.sdi_infor.progressive == 0x01 or self.sdi_infor.levelb == 0x01 then
				format_name = hAct .. "x" .. vAct0 .. "@" .. frameRate .. " Hz"
			else
				frameRate = frameRate * 2
				rate_n = rate_n * 2
				format_name = (hAct .. "x" .. (vAct0 + vAct1) .. "@" .. frameRate .. "i")
			end

			-- Check the recorded format info.
			local v = self.nostandardFormatInfo[format_name]
			if not v then	--Create this format information
				v = {
					formatId = format_name,
					formatName = format_name,
					frameRate = frameRate,
					hTotal = hTotal,
					hAct = hAct,
					vTotal0 = vTotal0,
					vAct0 = vAct0,
					vTotal1 = vTotal1,
					vAct1 = vAct1,
					interlaced = interlaced,
					rateMode = video_mode,
					tmdsFreq = pxclk,
					pixelRepeat = false,
					aspect_ratio = 0, --That is ASPECT_RATIO_AUTO
					video_system = 0 --That is AMBA_VIDEO_SYSTEM_AUTO
				}
				self.nostandardFormatInfo[format_name] = v
			end

			return self:_RecordVideoFormatInfo( "SDI", v )
		end
	end

	------------------------------------------------------

	FPGA_Rx:Reset()

	return FPGA_Rx
end --END of CreateFPGA_Rx()

return CreateFPGA_Rx
