local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local I2C_REG = require "chips.i2c_reg"

require "ndi_decoding"

--// Video format (0-SD, 1-1280x720, 2-1920x1035, 3-1920x1080)
--// Frame format (0,1,,7 ---> res,24p,25p,30p,50i,60i,50p,60p)

local VID_1920x1080 = 3
local VID_1920x1035 = 2
local VID_1280x720 	= 1
local VID_SD 		= 0

local RATE_P60 		= 7
local RATE_P50 		= 6
local RATE_I60 		= 5
local RATE_I50 		= 4
local RATE_P30 		= 3
local RATE_P25 		= 2
local RATE_P24 		= 1
local RATE_RES 		= 0


local NTSC_CLK 		= 1
local DEFAULT_CLK 	= 0

local formatMap = {
	-- 3G formats
	["1080p60"]		=	{	rate = RATE_P60,		fmt = VID_1920x1080,	clk = DEFAULT_CLK	},
	["1080p59"]		=	{	rate = RATE_P60,		fmt = VID_1920x1080,	clk = NTSC_CLK		},
	["1080p50"]		=	{	rate = RATE_P50,		fmt = VID_1920x1080,	clk = DEFAULT_CLK	},
	["1080p30"]		=	{	rate = RATE_P30,		fmt = VID_1920x1080,	clk = DEFAULT_CLK	},
	["1080p29"]		=	{	rate = RATE_P30,		fmt = VID_1920x1080,	clk = NTSC_CLK		},
	["1080p25"]		=	{	rate = RATE_P25,		fmt = VID_1920x1080,	clk = DEFAULT_CLK	},
	["1080p24"]		=	{	rate = RATE_P24,		fmt = VID_1920x1080,	clk = DEFAULT_CLK	},
	["1080p23"]		=	{	rate = RATE_P24,		fmt = VID_1920x1080,	clk = NTSC_CLK		},
	["1080i60"]		=	{	rate = RATE_I60,		fmt = VID_1920x1080,	clk = DEFAULT_CLK	},
	["1080i59"]		=	{	rate = RATE_I60,		fmt = VID_1920x1080,	clk = NTSC_CLK		},
	["1080i50"]		=	{	rate = RATE_I50,		fmt = VID_1920x1080,	clk = DEFAULT_CLK	},
	["720p60"]		=	{	rate = RATE_P60,		fmt = VID_1280x720,		clk = DEFAULT_CLK	},
	["720p59"]		=	{	rate = RATE_P60,		fmt = VID_1280x720,		clk = NTSC_CLK		},
	["720p50"]		=	{	rate = RATE_P50,		fmt = VID_1280x720,		clk = DEFAULT_CLK	},
	["720p30"]		=	{	rate = RATE_P30,		fmt = VID_1280x720,		clk = DEFAULT_CLK	},
	["720p29"]		=	{	rate = RATE_P30,		fmt = VID_1280x720,		clk = NTSC_CLK		},
	["720p25"]		=	{	rate = RATE_P25,		fmt = VID_1280x720,		clk = DEFAULT_CLK	},
	["720p24"]		=	{	rate = RATE_P24,		fmt = VID_1280x720,		clk = DEFAULT_CLK	},
	["720p23"]		=	{	rate = RATE_P24,		fmt = VID_1280x720,		clk = NTSC_CLK		},
	["480i60"]		=	{	rate = RATE_I60,		fmt = VID_SD,			clk = DEFAULT_CLK	},
	["576i50"]		=	{	rate = RATE_I50,		fmt = VID_SD,			clk = DEFAULT_CLK	},
	["525i"]		=	{	rate = RATE_I60,		fmt = VID_SD,			clk = DEFAULT_CLK	},
	["625i"]		=	{	rate = RATE_I50,		fmt = VID_SD,			clk = DEFAULT_CLK	}
}

local function CreateFPGA_Tx(i2c_dev, rxAddr)
-------------------------------------------------------
--{ The constructor wrap function.
-------------------------------------------------------

local _M = {
	first_run = true
}

i2c_dev = i2c_dev or "/dev/i2c-1"
	
local BASE_MAP = rxAddr or 0x69

_dbg.Info("FPGA_Rx", "Creating FPGA_Rx control object at 0x" .. string.format( "%02x", BASE_MAP) .. "@" .. i2c_dev .. "..." )

_M.i2c_reg = I2C_REG.Create()
if not _M.i2c_reg:Open( i2c_dev ) then
	return nil, "Fail to open I2C device"
end

-------------------------------------------------------
-- Object method: Register read
-------------------------------------------------------
function _M:ReadReg( map, addr, itemArr )
	return self.i2c_reg:ReadReg(map, addr, itemArr)
end

------------------------------------------------------
-- Register write
------------------------------------------------------
function _M:WriteReg( map, addr, itemArr )
	return self.i2c_reg:WriteReg(map, addr, itemArr)
end

----
function _M:Init()
	--TODO: nothing
	return true
end
----
function _M:Step()
	--TODO: nothing
end
----
function _M:UpdateVideoFormat( fmtInfo )
	if type(fmtInfo) ~= "table" or not fmtInfo.id then
		_dbg.Error("SDI-FPGA", "Invalid format information")
		return false
	end
	_dbg.Info("SDI-FPGA", "Configure current format id:", fmtInfo.id )
	local fmtSet = formatMap[fmtInfo.id]
	if not fmtSet then
		_dbg.Error("SDI-FPGA", "Unsupported format ID:", fmtInfo.id, ", configuration is ignored." )
		return false
	end
	local regVal = (fmtSet.fmt << 14) | (fmtSet.rate << 11) | (fmtSet.clk << 10)
	local fpgaVal = 0x80+(fmtSet.clk<<6)+(fmtSet.fmt<<4)+(fmtSet.rate<<1)
	local regVal0

	if fmtInfo.deInterlace then
		regVal0 = regVal
		regVal = regVal | 0x80
		fpgaVal = fpgaVal |0x04
	else
		regVal0 = regVal | 0x80
	end

	_dbg.Info("SDI-FPGA", "Format register value is:", string.format("%08x", regVal))
	_dbg.Info("SDI-FPGA", "Format fpga value is:", string.format("%08x", fpgaVal))
	local res = self:WriteReg( BASE_MAP,0xA2,{fpgaVal} )
	if not res then
		local t = 0
		while t<10 do
			if res then
				break
			end
			res = self:WriteReg( BASE_MAP,0xA2,{fpgaVal} )
			coaf.Sleep(1)
			t = t +1
		end
	end
	if fmtInfo.format_mode == "SD" then
		coaf.Sleep(0.5)
	end
	NDI.DECODING.write_reg( 0x280, 0x0e, regVal )
	

	--A temporary FPGA bug fix.
	if fmtInfo.interlaced and _M.first_run then
		_M.first_run = false
		coaf.Sleep(0.1)
		NDI.DECODING.write_reg( 0x280, 0x0e, regVal0 )
		coaf.Sleep(0.1)
		NDI.DECODING.write_reg( 0x280, 0x0e, regVal )
	end

	return true
end

----

function _M:UpdateAudioFormat( fmtInfo )
	local channels = 2
	if fmtInfo and fmtInfo.no_channels then
		channels = fmtInfo.no_channels
	end
	if channels%4==0 then
		channels = channels//4
	else
		channels = (channels//4)+1
	end
	_dbg.Info("SDI-FPGA", "Format audio format value is:", string.format("%08x", 2^channels-1))
	local res = self:WriteReg( BASE_MAP,0xA8,{2^channels-1} )
	if not res then
		local t = 0
		while t<10 do
			if res then
				break
			end
			res = self:WriteReg( BASE_MAP,0xA8,{2^channels-1} )
			coaf.Sleep(1)
			t = t +1
		end
	end
	--TODO: nothing.
	return true
end

_M:Init()

return _M

-------------------------------------------------------
--} End of the constructor wrap function.
-------------------------------------------------------
end

return CreateFPGA_Tx
