local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local I2C_REG = require "chips.i2c_reg"

local _M = {
	--Store all I2C transfer objects, index based on I2C device.
	i2c_trans = {}
}

-------------------------------------------------------------
-- The PIN
-------------------------------------------------------------
local pin_meta = {}

function pin_meta:read()
	local reg_offs = math.floor((self.pin_index-1)/8)
	local v, msg = self.chip.i2c_trans:ReadReg(self.chip.address, reg_offs, {0})
	if type(v) == "table" then
		return (v[1] & (1<<reg_offs)) ~= 0
	else
		return false
	end
end

function pin_meta:write(val)
	self.value = val
	self:update(true)
end

function pin_meta:update(no_setting)
	local reg_offs = math.floor((self.pin_index-1)/8)
	local pin_offs = (self.pin_index-1)%8
	if not no_setting then
		local reg = self.chip.io_cfgs[reg_offs+1]
		if self.direction == "in" then
			reg = reg | (1<<pin_offs)
		else
			reg = reg & (~(1<<pin_offs))
		end
		self.chip.io_cfgs[reg_offs+1] = reg
		self.chip.i2c_trans:WriteReg(self.chip.address, 0x0c + reg_offs, {reg})

		reg = self.chip.polarity_cfgs[reg_offs+1]

		if self.revert then
			reg = reg | (1<<pin_offs)
		else
			reg = reg & (~(1<<pin_offs))
		end
		self.chip.polarity_cfgs[reg_offs+1] = reg
		self.chip.i2c_trans:WriteReg(self.chip.address, 0x08 + reg_offs, {reg})
	end

	if self.direction == "out" then
		local val = self.value
		if val == nil then return end
		
		if self.revert then
			val = not val
		end
		local reg = self.chip.outs[reg_offs+1]
		if val then
			reg = reg | (1<<pin_offs)
		else
			reg = reg & (~(1<<pin_offs))
		end
		self.chip.outs[reg_offs+1] = reg
		self.chip.i2c_trans:WriteReg(self.chip.address, 0x04 + reg_offs, {reg})
	end
end

pin_meta.__index = pin_meta

-------------------------------------------------------------
-- The CHIP
-------------------------------------------------------------
local chip_meta = {
}

function chip_meta:Pin(idx, direction, revert)
	if not idx or idx < 1 or idx > 24 then
		_dbg.Error("TCA6424A", "Invalid PIN ", idx)
		return nil
	end

	revert = revert and true or false

	direction = direction or "out"
	direction = direction:lower()
	local init_val
	if direction == "high" then
		direction = "out"
		init_val = true
	elseif direction == "low" then
		direction = "out"
		init_val = false
	elseif direction ~= "out" and direction ~= "in" then
		_dbg.Error("TCA6424A", "Bad PIN direction of '", direction, "', should be in/out/low/high." )
		return nil
	end

	local pin_obj = self.pin_objs[idx]
	if pin_obj then
		if pin_obj.direction ~= direction or (init_val ~= nil and pin_obj.value ~= init_val) or (revert ~= pin_obj.revert) then
			_dbg.Warn("TCA6424A", "Changed the PIN settings to dir=", direction, ",init=", init_val, ",revert=", revert)
			pin_obj.direction = direction
			if init_val ~= nil then
				pin_obj.value = init_val
			end
			pin_obj.revert = revert
			pin_obj:update()
		end
		return pin_obj
	end

	pin_obj = {
		chip = self,
		pin_index = idx,
		direction = direction,
		value = init_val,
		revert = revert
	}

	setmetatable(pin_obj, pin_meta)
	self.pin_objs[idx] = pin_obj
	pin_obj:update()

	return pin_obj
end

function chip_meta:Reset()
	--NOTE: For burst transmission, highest bit should be 1.
	self.i2c_trans:WriteReg(self.address, 0x0c + 0x80, {table.unpack(self.io_cfgs)})
	self.i2c_trans:WriteReg(self.address, 0x08 + 0x80, {table.unpack(self.polarity_cfgs)})
	self.i2c_trans:WriteReg(self.address, 0x04 + 0x80, {table.unpack(self.outs)})
end

chip_meta.__index = chip_meta
-------------------------------------------------------------

--Create a chip:
function _M.Chip(i2c_dev, i2c_addr)
	i2c_addr = tonumber(i2c_addr)
	if not i2c_dev or not i2c_addr then
		_dbg.Error("TCA6424A", "Invalid I2C device or device address!")
		return nil
	end

	local i2c_trans = _M.i2c_trans[i2c_dev]
	if not i2c_trans then
		i2c_trans = I2C_REG.Create()
		if not i2c_trans:Open(i2c_dev) then
			_dbg.Error("TCA6424A", "Open I2C device failed! device=", i2c_dev, ", address=", string.format("%02X",i2c_addr))
			return nil
		end
		_dbg.Info("TCA6424A", "Opened GPIO controller chip at device=", i2c_dev, ", address=", string.format("%02X",i2c_addr))
		_M.i2c_trans[i2c_dev] = i2c_trans
	else
		_dbg.Info("TCA6424A", "Using exist I2C transfer object for ", i2c_dev, ", address=", string.format("%02X",i2c_addr) )
	end

	local chip_obj = {
		i2c_trans = i2c_trans,
		address = i2c_addr,
		io_cfgs = {0x00, 0x00, 0x00},
		polarity_cfgs = {0x00, 0x00, 0x00},
		outs = {0xff, 0xff, 0xff},
		pin_objs = {}
	}

	setmetatable(chip_obj, chip_meta)
	return chip_obj
end

return _M
