local gioperiphery = require "gioperiphery"
local SPI = gioperiphery.SPI

local _M = {
	sleep = nil
}

function _M.Sleep(nSec)
	if _M.sleep then
		_M.sleep(nSec)
	end
end

local WREN = 0x06	--Write enable.
local WRDI = 0x04	--Write disable.

local RDSR = 0x05    --读状态寄存器
local RDSR2 = 0x35   --读状态寄存器 只有华邦spi nor flash
local WRSR = 0x01    --写状态寄存器，WEL为0不允许写
local READ = 0x03    --读字节
local FAST_READ = 0x0B    --快读指令
local PP = 0x02           --页写指令
local SE = 0xD8           --区域擦除
local BE = 0xC7           --批擦除
local DP = 0xB9           --深度掉电模式
local RES = 0xAB          --从深度掉电模式苏醒
local UIDN = 0x4B         --flash芯片的uuid
local PS = 0x42          --安全编程寄存器
local RS = 0x48			 --安全读取寄存器

local chip_meta = {}

function chip_meta:WriteEnable()
	self.spi:transfer({WREN})
end

function chip_meta:WriteDisable()
	self.spi:transfer({WRDI})
end

function chip_meta:ReadSR()
	local ret = self.spi:transfer({RDSR,0xff})
	if ret then
		return ret[2]
	else
		return nil
	end
end

function chip_meta:ReadSR2()
	local ret = self.spi:transfer({RDSR2,0xff})
	if ret then
		return ret[2]
	else
		return nil
	end
end

function chip_meta:WriteSR(val)
	self:WriteEnable()
	self.spi:transfer({WRSR, val})
	self:WriteDisable()
end

function chip_meta:Read(addr,size)
	local rTable = {READ, (addr>>16)&0xff, (addr>>8)&0xff, addr&0xff}
	local i
	for i=1, size do
		table.insert(rTable,0xff)
	end
	rTable = self.spi:transfer(rTable)
	if rTable then
		table.remove(rTable,1)
		table.remove(rTable,1)
		table.remove(rTable,1)
		table.remove(rTable,1)
	end
	return rTable
end

function chip_meta:Write(addr,buf)
	if type(buf) == "string" then
		local wTable = { string.byte(buf,1,#buf) }
		buf = wTable
	elseif type(buf) ~= "table" then
		return false
	end
	table.insert(buf,1,PP)
	table.insert(buf,2,(addr>>16)&0xff)
	table.insert(buf,3,(addr>>8)&0xff)
	table.insert(buf,4,addr&0xff)
	self:WriteEnable()
	self.spi:transfer(buf)
	rb = self:ReadSR()
	while (rb & 0xff) ~= 0 do
		_M.Sleep(0.001)
		rb = self:ReadSR()
	end
	self:WriteDisable()
	return true
end

function chip_meta:EraseSector(addr)
	self:WriteEnable()
	self.spi:transfer({SE, (addr>>16)&0xff, (addr>>8)&0xff, addr&0xff})
	rb = self:ReadSR()
	while (rb & 0xff) ~= 0 do
		_M.Sleep(0.001)
		rb = self:ReadSR()
	end
	self:WriteDisable()
end

function chip_meta:EraseBulk()
	self:WriteEnable()
	self.spi:transfer({BE})
	rb = self:ReadSR()
	while (rb & 0xff) ~= 0 do
		_M.Sleep(0.001)
		rb = self:ReadSR()
	end
	self:WriteDisable()
end

function chip_meta:WriteSecurity()
	if type(buf) == "string" then
		local wTable = { string.byte(buf,1,#buf) }
		buf = wTable
	elseif type(buf) ~= "table" then
		return false
	end
	table.insert(wTable,1,PS)
	table.insert(wTable,2,(addr>>16)&0xff)
	table.insert(wTable,3,(addr>>8)&0xff)
	table.insert(wTable,4,addr&0xff)
	self:WriteEnable()
	self.spi:transfer(wTable)
	rb = self:ReadSR()
	while (rb & 0xff) ~= 0 do
		_M.Sleep(0.001)
		rb = self:ReadSR()
	end
	self:WriteDisable()
	return true
end

function chip_meta:ReadSecurity()
	local rTable = {RS, (addr>>16)&0xff, (addr>>8)&0xff, addr&0xff}
	local i
	for i=1, size do
		table.insert(rTable,0xff)
	end
	rTable = self.spi:transfer(rTable)
	if rTable then
		table.remove(rTable,1)
		table.remove(rTable,1)
		table.remove(rTable,1)
		table.remove(rTable,1)
	end
	return rTable
end

function chip_meta:GetUUID()
	local wrBuf = {UIDN, 0, 0, 0, 0, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff}
	local rBuf = self.spi:transfer(wrBuf)
	local uuid
	if rBuf then
		uuid = string.format("%02X%02X%02X%02X%02X%02X%02X%02X", rBuf[6], rBuf[7], rBuf[8], rBuf[9], rBuf[10], rBuf[11], rBuf[12],rBuf[13])
	end
	return uuid
end

function chip_meta:ProgramFile(filename,veri)
	local uuid = self:GetUUID()
	if not uuid then
		return false, "Fail to check chip UUID!"
	end

	print( "M25P80: Program binary file '" .. (filename or "<nil>") .. "' to chip:", uuid )
	local file = io.open(filename,"rb")
	if not file then
		print("M25P80: ERROR! Fail to open binary file '" .. (filename or "<nil>") .. "'" )
		return false, "Fail to open binary file of " .. (filename or "<nil>")
	end

	print( "M25P80: Erasing the chip..." )
	self:EraseBulk()
	print( "M25P80: Programing with sector size 256..." )
	local blkCnt = 0
	local fileSize = 0
	local blk = file:read(256)
	local veriStatus
	while blk do
		veriStatus = true
		fileSize = fileSize + #blk
		self:Write(blkCnt*256, blk)
		if veri then
			local readback = self:Read(blkCnt*256,256)
			local i
			for i=1, #blk do
				if blk:byte(i) ~= readback[i] then
					veriStatus = false
					break
				end
			end
		end
		blkCnt = blkCnt + 1
		print( "  .. Write block " .. blkCnt .. ": " .. (veriStatus and "OK" or "VERIFY FAIL") )
		blk = file:read(256)
	end
	file:close()
	print( "M25P80: Program done, file size:", fileSize )
	return true
end

function chip_meta:VerifyFile(filename)
	local uuid = self:GetUUID()
	if not uuid then
		return false, "Fail to check chip UUID!"
	end

	print( "M25P80: Verify binary file '" .. (filename or "<nil>") .. "' with chip:", uuid )
	local file = io.open(filename,"rb")
	if not file then
		print("M25P80: ERROR! Fail to open binary file '" .. (filename or "<nil>") .. "'" )
		return false, "Fail to open binary file of " .. (filename or "<nil>")
	end

	local blkCnt = 0
	local fileSize = 0
	local blk = file:read(256)
	local veriStatus
	local veriFail = false
	while blk do
		veriStatus = nil
		local readback = self:Read(blkCnt*256,256)
		local i
		for i=1, #blk do
			if blk:byte(i) ~= readback[i] then
				veriStatus = i
				veriFail = true
				break
			end
		end
		print( "Block# " .. blkCnt .. ", start address: ", string.format("%06X", blkCnt*256), veriStatus and (", verify fail at offset " .. (veriStatus-1)) or ", verify OK" ) 
		fileSize = fileSize + #blk
		blkCnt = blkCnt + 1
		blk = file:read(256)
	end
	file:close()
	print( "Total file size:", fileSize )
	return not veriFail
end

function chip_meta:DownloadFile(filename,sizeOpt)
	if not sizeOpt then
		sizeOpt = "1M"
	end

	local size_num, size_unit = sizeOpt:match("([0-9%.]+)([kKmM]*)" )
	if not size_num then
		print( "M25P80: ERROR: Download file but with invalid size:", sizeOpt )
		return false, "Invalid size"
	end

	size_num = tonumber(size_num)
	if size_unit == 'k' or size_unit == 'K' then
		size_num = size_num * 1024
	elseif size_unit == 'm' or size_unit == 'M' then
		size_num = size_num * 1024 * 1024
	end

	size_num = math.ceil(size_num)
	size_num = (size_num + 255) & 0xfffff00

	local uuid = self:GetUUID()
	if not uuid then
		return false, "Fail to check chip UUID!"
	end

	print( "M25P80: Download flashed binary file '" .. (filename or "<nil>") .. "' with chip:", uuid )
	print( "M25P80: Download size:", size_num )
	local file = io.open(filename,"wb")
	if not file then
		print("M25P80: ERROR! Fail to open binary file '" .. (filename or "<nil>") .. "' for saving." )
		return false, "Fail to open binary file of " .. (filename or "<nil>")
	end

	local blkCnt = 0
	local fileSize = 0
	while blkCnt < (size_num/256) do
		veriStatus = nil
		local readback = self:Read(blkCnt*256,256)
		print( "READ Block# " .. blkCnt .. ", start address: ", string.format("%06X", blkCnt*256) ) 
		fileSize = fileSize + 256
		file:write( string.char( table.unpack(readback) ) )
		blkCnt = blkCnt + 1
	end
	file:close()
	print( "Total file size:", fileSize )
	return true
end

chip_meta.__index = chip_meta

---------------------------------------

function _M.createNew( pinCfg )
	local spi = SPI( pinCfg, 0, 0 )
	if not spi then
		return nil, "Fail to create GPIO controller for M25P80"
	end
	return setmetatable({
		spi = spi
	}, chip_meta )
end

return _M

