#!/usr/bin/env lua
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

local curdir = coaf.GetCwd()
coaf.ChDir(coaf.GetAppDir(arg))

_G.print = function(...)
	_dbg.Print( "FPGA-PROG", ... )
end

local peri = require "periphery"
local GPIO = peri.GPIO
local m25p80 = require "chip_m25p80"

--Change back to current dir.
coaf.ChDir(curdir)

m25p80.sleep = coaf.Sleep

--Advanced: so you can configure the FPGA flash pins.
local pinmux_ctrl = {
	sw = { pin=983 },	--GPIO76: 1 - SPI to flash; 0 - FPGA to flash;
	clk = { pin=987 },  	--SPI CLK;
	din = { pin=985 },  --SPI DIN;(Aka Flash DOUT)
	dout = { pin=984 },  --SPI DOUT;(Aka Flash DIN)
	cs = { pin=986 }  	--SPI CSN
}

local function InitPinmux()
	if type(pinmux_ctrl.sw) == "table" and pinmux_ctrl.sw.pin then
		pinmux_ctrl.sw.gpio = GPIO(pinmux_ctrl.sw.pin, "low")
	else
		pinmux_ctrl.sw = nil
	end
end

local verify = false
local download = false
local downsize = "1M"

local file = nil
if arg[1] == "--help" then
	_dbg.Print( "FPGA-PROG", "Usage: program_fpga.lua [--verify|-v] [--download|-d <size>] <filename>" )
	os.exit(0)
elseif arg[1] == "-v" or arg[1] == "--verify" then
	verify = true
	file = arg[2]
elseif arg[1] == '-d' or arg[1] == "--download" then
	download = true
	downsize = arg[2]
	file = arg[3]
else
	file = arg[1]
end

if not file then
	_dbg.Print( "FPGA-PROG", "Usage: program_fpga.lua [--verify|-v] [--download|-d <size>] <filename>" )
	os.exit(0)
end

coaf.Init("progfpga")

local function Main(args)
	_dbg.Info( "FPGA-PROG", "Initialize pinmux..." )
	InitPinmux()
	if not pinmux_ctrl.clk or not pinmux_ctrl.cs or not pinmux_ctrl.din or not pinmux_ctrl.dout then
		_dbg.Error( "FPGA-PROG", "Invalid FPGA flash SPI pins configured!" )
		os.exit(1)
	end

	local pinOpts = "clk=" .. pinmux_ctrl.clk.pin .. "," ..
					"cs=" .. pinmux_ctrl.cs.pin .. "," ..
					"din=" .. pinmux_ctrl.din.pin .. "," ..
					"dout=" .. pinmux_ctrl.dout.pin
	
	_dbg.Info( "FPGA-PROG", "Create M25P80 chip instance...", pinOpts )
	local chip = m25p80.createNew( pinOpts )
	if not chip then
		_dbg.Error( "FPGA-PROG", "Fail to create M25P80 chip instance" )
		os.exit(1)
	end
	
	_dbg.Info( "FPGA-PROG", "Switch FPGA flash to CPU for programming..." )
	if pinmux_ctrl.sw then
		pinmux_ctrl.sw.gpio:write(true)
	end

	_dbg.Info( "FPGA-PROG", "Check M25P80 chip's UUID..." )
	--coaf.Sleep(3)
	local uuid = chip:GetUUID()
	if not uuid then
		_dbg.Info( "FPGA-PROG", "Switch FPGA flash back to FPGA." )
		if pinmux_ctrl.sw then
			pinmux_ctrl.sw.gpio:write(false)
		end
		_dbg.Error( "FPGA-PROG", "Fail to get chip UUID!" )
		os.exit(1)
	else
		_dbg.Info( "FPGA-PROG", "Got M25P80 UUID:", uuid )
	end
	
	if verify then
		_dbg.Info( "FPGA-PROG", "Verify chip compares to file:", file, "..." )
		if chip:VerifyFile(file) then
			_dbg.Info( "FPGA-PROG", "Chip verify DONE." )
		else
			_dbg.Error( "FPGA-PROG", "Chip verify FAIL!" )
		end
	elseif download then
		_dbg.Info( "FPGA-PROG", "Download chip content to file:", file, ", size=", downsize, "..." )
		if chip:DownloadFile(file, downsize) then
			_dbg.Info( "FPGA-PROG", "Chip download DONE." )
		else
			_dbg.Error( "FPGA-PROG", "Chip download FAIL!" )
		end
	else
		_dbg.Info( "FPGA-PROG", "Program binary file:", file, "..." )
		if chip:ProgramFile(file,true) then
			_dbg.Info( "FPGA-PROG", "Chip programmed DONE." )
		else
			_dbg.Error( "FPGA-PROG", "Chip programmed FAIL!" )
		end
	end
	
	_dbg.Info( "FPGA-PROG", "Switch FPGA flash back to FPGA." )
	if pinmux_ctrl.sw then
		pinmux_ctrl.sw.gpio:write(false)
	end

	--[[
	if not verify and not download and pinmux_ctrl.reset then
		_dbg.Info( "FPGA-PROG", "Reset the FPGA..." )
		--HIMM.write( pinmux_ctrl.reset.pingroup + (1<<(pinmux_ctrl.reset.pin+2)), 0 << pinmux_ctrl.reset.pin )
		coaf.Sleep(0.5)
		--HIMM.write( pinmux_ctrl.reset.pingroup + (1<<(pinmux_ctrl.reset.pin+2)), 1 << pinmux_ctrl.reset.pin )
	end
	--]]

	coaf.Exit(0)
end

coaf.Start(Main)

os.exit(coaf.CoAF_returnCode or 0)

